import os
import random
import yaml
from pathlib import Path
from typing import Optional, Dict, Any, TypeVar, Type, cast
import importlib.resources as resources
from dotenv import load_dotenv
import importlib
from typing import Callable
from pydantic import BaseModel
from deprecated import deprecated


class SchemaValidationError(ValueError):
    pass

VALIDATION_EXCEPTIONS = ["options"]
VALIDATION_EXTRAS = ["package", "function"]

def resolve_agent_schema(agent_schema: str | Path | dict):
    """
    Resolve the agent schema from a string, path (to yaml file) or dict.
    """
    if isinstance(agent_schema, str):
        agent_schema = Path(agent_schema)
    if isinstance(agent_schema, Path):
        if not agent_schema.exists():
            raise ValueError(
                f"In constructor agent_schema path  does not exist: ({str(agent_schema.absolute())})!")
        with open(agent_schema) as f:
            agent_schema = yaml.safe_load(f)
    if not isinstance(agent_schema, dict):
        raise ValueError(
            f"In constructor agent_schema parameter should be string, Path or dict!")

    return agent_schema

def resolve_and_validate_agent_schema(agent_schema: str | Path | dict | None, default_file_name: str):
    reference_schema = resolve_agent_schema(Path(Path(__file__).parent, "config", default_file_name))
    if agent_schema is None:
        return reference_schema

    agent_schema = resolve_agent_schema(agent_schema)
    validate_schema(reference_schema, agent_schema)

    return agent_schema


def create_fields_set(source: dict[str, Any], fields_set: set[str]):
    for key in source:
        fields_set.add(key)
        if (key not in VALIDATION_EXCEPTIONS) and isinstance(source[key], dict):
            create_fields_set(source[key], fields_set)


def validate_schema(reference: dict[str, Any], schema: dict[str, Any]):
    reference_set: set[str] = set(VALIDATION_EXTRAS)
    schema_set: set[str] = set()
    create_fields_set(reference, reference_set)
    create_fields_set(schema, schema_set)
    error_fields = []
    for field in schema_set:
        if field not in reference_set and field not in VALIDATION_EXCEPTIONS:
            error_fields.append(field)

    if len(error_fields) > 0:
        raise SchemaValidationError(f" Fields {error_fields} not exists in yaml schema. Choose from {reference_set}")


def resolve_llm_options(agent_schema: dict, llm_options: dict):
    if llm_options is None:
        llm_options = agent_schema.get("options", None)
    if llm_options is None:
        raise ValueError(
            "llm_options should not be None. You should pass it through llm_options or agent_schema['options'].")

    return llm_options


def resolve_system_prompt(agent_schema: dict):
    system_prompt = agent_schema.get("system_prompt", None)
    system_prompt_path = agent_schema.get("system_prompt_path", None)
    if (system_prompt is not None) and (system_prompt_path is not None):
        raise ValueError("You should use only one of system_prompt or system_prompt_path not both together.")
    if system_prompt_path is not None:
        path = Path(system_prompt_path)
        if path.exists():
            system_prompt = path.read_text(encoding="utf8")
    return system_prompt


def resolve_tools(agent_schema: dict) -> list[Callable]:
    function_list:list[Callable] = []
    tools = agent_schema.get('tools', None)
    if tools is None:
        return None
    for entry in tools:
        package_name: str = entry['package']
        function_name: str = entry['function']
        try:
            # Dynamically import the package
            package = importlib.import_module(package_name)
            # Get the function from the package
            func = getattr(package, function_name)
            function_list.append(func)
        except (ImportError, AttributeError) as e:
            print(f"Error importing {function_name} from {package_name}: {e}")

    return function_list


def rotate_env_keys() -> str:
    load_dotenv()
    keys = []
    index = 1

    while True:
        # Determine the environment variable name
        key_name = 'KEY' if index == 0 else f'KEY_{index}'
        key_value = os.getenv(key_name)

        # Break the loop if the key is not found
        if key_value is None:
            break

        # Add the found key to the list
        keys.append(key_value)
        index += 1

    # Raise an error if no keys are found
    if not keys:
        raise ValueError("No keys found in environment variables")

    # Randomly choose one of the available keys
    return random.choice(keys)

def load_config(resource: str, package: str = "just_agents.simple.config") -> Dict[str, Any]:
    """
    :rtype: yaml config
    """
    if Path(resource).exists():
        return yaml.safe_load(Path(resource).open("r"))
    in_config = Path("config") / resource
    if in_config.exists():
        return yaml.safe_load( (Path("config") / resource).open("r") )
    else:
        # Load from package resources
        with resources.open_text(package, 'agent_prompts.yaml') as file:
            return yaml.safe_load(file)


@deprecated(version='0.3.0', reason="You should use native serialization of BaseAgent or its descendants")
def build_agent(agent_schema: str | Path | dict):
    from just_agents.simple.cot_agent import ChainOfThoughtAgent
    from just_agents.simple.llm_session import LLMSession
    agent_schema = resolve_agent_schema(agent_schema)
    class_name = agent_schema.get("class", None)
    if class_name is None or class_name == "LLMSession":
        return LLMSession(agent_schema=agent_schema)
    elif class_name == "ChainOfThoughtAgent":
        return ChainOfThoughtAgent(agent_schema=agent_schema)


######### Pydantic models ###########

BaseT = TypeVar('BaseT', bound=BaseModel)
def extract_common_fields(selected_class: Type[BaseT], instance: BaseModel) -> BaseT:
    """
    Trims and typecasts an instance of a class to only include the fields of the selected class.

    :param selected_class: The class type to trim to.
    :param instance: The instance of the class to be trimmed.
    :return: An instance of the selected class populated with the relevant fields from the provided object.
    """
    # Extract only the fields defined in the base class
    base_fields = {field: getattr(instance, field) for field in selected_class.model_fields}

    # Instantiate and return the base class with these fields
    return selected_class(**base_fields)

def trim_to_parent(instance: BaseT) -> BaseModel:
    """
    Trims an instance of a derived class to only include the fields of its direct parent class.
    :param instance: The instance of the derived class to be trimmed.
    :return: An instance of the parent class populated with the relevant fields from the derived class.
    """
    # Get the direct parent class of the instance
    parent_class = type(instance).__class__.__bases__[0]

    # Instantiate and return the parent class with these fields
    parent_instance = extract_common_fields(parent_class, instance)
    return cast(parent_class, parent_instance)




from typing import Any
from abc import ABC, abstractmethod

class AbstractStreamingProtocol(ABC):
    @abstractmethod
    def get_chunk(self, index:int, delta:str, options:dict) -> Any:
        raise NotImplementedError("You need to implement get_chunk() first!")

    @abstractmethod
    def done(self) -> str:
        raise NotImplementedError("You need to implement done() first!")



from enum import Enum
from just_agents.streaming.protocols.interfaces.IProtocolAdapter import IProtocolAdapter, ExecuteToolCallback

class StreamingMode(str, Enum):
    openai = "openai"
    qwen2 = "qwen2"

    def __new__(cls, value, *args, **kwargs):
        obj = str.__new__(cls, value)
        obj._value_ = value
        return obj

    def __str__(self):
        return str(self.value)

class ProtocolAdapterFactory:
    @staticmethod
    def get_protocol_adapter(
            mode: StreamingMode,
            execute_functions: ExecuteToolCallback,

    ) -> IProtocolAdapter:
        if mode == StreamingMode.openai:
            from just_agents.streaming.openai_protocol_adapter import OAIAdapter
            return OAIAdapter(
                execute_function_hook=execute_functions,
            )
        elif mode == StreamingMode.qwen2:
            # todo: implement
            raise NotImplementedError("Qwen2 streaming is not yet implemented for pydantic-based agents")
        else:
            raise ValueError("Unknown streaming method")





from just_agents.core.interfaces.IAgent import *

# New TypeVar for Thought
THOUGHT_TYPE = TypeVar('THOUGHT_TYPE', bound='IThought')

class IThought(BaseModel):
    @abstractmethod
    def is_final(self) -> bool:
        raise NotImplementedError("You need to implement is_final() abstract method first!")


class IThinkingAgent(
    IAgent[AbstractQueryInputType, AbstractQueryResponseType, AbstractStreamingChunkType],
    Generic[AbstractQueryInputType, AbstractQueryResponseType, AbstractStreamingChunkType, THOUGHT_TYPE]
):
    
    @abstractmethod
    def thought_query(self, response: AbstractQueryInputType, **kwargs) -> THOUGHT_TYPE:
        raise NotImplementedError("You need to implement thought_query abstract method first!")

    def think(self, 
              query: AbstractQueryInputType, 
              max_iter: int = 3, 
              chain: Optional[list[THOUGHT_TYPE]] = None,
              **kwargs  ) -> tuple[Optional[THOUGHT_TYPE], Optional[list[THOUGHT_TYPE]]]:
        """
        This method will continue to query the agent until the final thought is not None or the max_iter is reached.
        Returns a tuple of (final_thought, thought_chain)
        """
        current_chain = chain or []
        response = self.query(query, **kwargs) 
        thought = self.thought_query(query, **kwargs) #queries itself with thought as expected output
        new_chain = [*current_chain, thought] #updates chain with the new thought
        if thought.is_final() or max_iter <= 0:
            return (self.thought_query(query, **kwargs), new_chain) #returns the final thought and the chain that preceded it
        else:
            return self.think(query, max_iter - 1, new_chain, **kwargs) #continues the thought process


from abc import ABC, abstractmethod
import ast
import json
from typing import Type, Union, Generator, AsyncGenerator, Any, TypeVar, Generic, List, Optional, Callable, Coroutine, Protocol, Tuple
from pydantic import BaseModel, ConfigDict
import re

# Add this near the top of the file, before any class definitions
import pydantic
pydantic.TypeAdapter.validate_python.__globals__['ConfigDict'] = ConfigDict(
    max_error_message_length=None
)

# Define generic types for inputs and outputs
Self = TypeVar("Self") # 3.11+ only, replacement for 3.8+ compatibility
AbstractQueryInputType = TypeVar("AbstractQueryInputType")

AbstractQueryResponseType = TypeVar("AbstractQueryResponseType")
AbstractStreamingChunkType = TypeVar("AbstractStreamingChunkType")

AbstractAgentInputType = TypeVar('AbstractAgentInputType', bound=BaseModel)
AbstractAgentOutputType = TypeVar('AbstractAgentOutputType', bound=BaseModel)

# Define the type that represents streaming responses
AbstractStreamingGeneratorResponseType = Union[
    Coroutine[Any, Any, AbstractQueryResponseType],
    Coroutine[Any, Any, AsyncGenerator[AbstractStreamingChunkType, None]],
    Generator[AbstractStreamingChunkType, None ,None],
    AsyncGenerator[AbstractStreamingChunkType, None]
]

# Signature for a query function
QueryFunction = Callable[[Self,AbstractQueryInputType,...],Any]
ResponseFunction = Callable[...,AbstractQueryResponseType]
StreamingResponseFunction = Callable[...,AbstractStreamingGeneratorResponseType]

# Signatures for listener templates
class QueryListener(Protocol[AbstractQueryResponseType]):
    def __call__(self, input_query: AbstractQueryResponseType, *args: Any, **kwargs: Any) -> None:
        ...

class ResponseListener(Protocol[AbstractQueryResponseType]):
    def __call__(self, result: AbstractQueryResponseType, *args: Any, **kwargs: Any) -> None:
        ...

class IAgent(ABC, Generic[AbstractQueryInputType, AbstractQueryResponseType, AbstractStreamingChunkType]):

    @abstractmethod
    def query(self, query_input: AbstractQueryInputType) -> Optional[AbstractQueryResponseType]:
        raise NotImplementedError("You need to implement query() abstract method first!")
    
    def query_structural(
        self, 
        query_input: AbstractQueryInputType, 
        parser: Type[BaseModel] = BaseModel,
        **kwargs
    ) -> Union[dict, BaseModel]:
        """
        Query the agent and parse the response according to the provided parser.
        Attempts multiple parsing strategies in the following order:
        1. Direct dict usage if already a dict
        2. Standard json parsing
        3. AST literal eval as final fallback
        """
        raw_response = self.query(query_input, **kwargs)

        # If already a dict, no parsing needed
        if isinstance(raw_response, dict):
            return raw_response if parser == dict else parser.model_validate(raw_response)

        if not isinstance(raw_response, str):
            raw_response = str(raw_response)

        # Check for and fix common double-escaping issues
        if '\\\\' in raw_response:
            raw_response = raw_response.replace('\\\\', '\\')

        parsing_errors = []
        
        # Try standard json first
        try:
            response_dict = json.loads(raw_response)
            return response_dict if parser == dict else parser.model_validate(response_dict)
        except json.JSONDecodeError as e:
            parsing_errors.append(f"Standard JSON parsing failed: {str(e)}")

        # Final fallback: AST literal eval
        try:
            response_dict = ast.literal_eval(raw_response)
            if isinstance(response_dict, dict):
                return response_dict if parser == dict else parser.model_validate(response_dict)
            parsing_errors.append("AST parsing succeeded but result was not a dict")
        except (ValueError, SyntaxError) as e:
            parsing_errors.append(f"AST literal_eval parsing failed: {str(e)}")

        # If all parsing attempts failed, raise detailed error
        raise ValueError(
            f"Failed to parse response using multiple methods:\n"
            f"{chr(10).join(parsing_errors)}\n"
            f"Original response:\n{raw_response}"
        )

    
    @abstractmethod
    def stream(self, query_input: AbstractQueryInputType) -> Optional[AbstractStreamingGeneratorResponseType]:
        raise NotImplementedError("You need to implement stream() abstract method first!")

# Define IAgentWithInterceptors with methods to manage interceptors
class IAgentWithInterceptors(
        IAgent[AbstractQueryInputType, AbstractQueryResponseType, AbstractStreamingChunkType],
        ABC,
        Generic[AbstractQueryInputType, AbstractQueryResponseType, AbstractStreamingChunkType]
    ):

    _on_query: List[QueryListener[AbstractQueryInputType]]
    _on_response: List[ResponseListener[AbstractQueryResponseType]]

    # Methods to manage on_query listeners
    def handle_on_query(self, input_query: AbstractQueryResponseType, *args, **kwargs) -> None:
        for handler in self._on_query:
            handler(input_query, *args, **kwargs)

    def add_on_query_listener(self, listener: QueryListener[AbstractQueryInputType]) -> None:
        self._on_query.append(listener)

    def remove_on_query_listener(self, listener: QueryListener[AbstractQueryInputType]) -> None:
        self._on_query.remove(listener)

    # Methods to manage on_response listeners
    def handle_on_response(self, query_result: AbstractQueryResponseType, *args, **kwargs) -> None:
        for handler in self._on_response:
            handler(query_result, *args, **kwargs)

    def add_on_response_listener(
            self,
            listener: ResponseListener[AbstractQueryResponseType]
    ) -> None:
        self._on_response.append(listener)

    def remove_on_response_listener(
            self,
            listener: ResponseListener[AbstractQueryResponseType]
    ) -> None:
        self._on_response.remove(listener)




from pydantic import BaseModel, Field, PrivateAttr
from typing import Optional, Callable, List, Dict, Union
from functools import singledispatchmethod
from just_agents.core.interfaces.IMemory import IMemory
from just_agents.core.types import Role, AbstractMessage, SupportedMessages, SupportedMessage
from litellm.types.utils import Function

OnMessageCallable = Callable[[AbstractMessage], None]
OnFunctionCallable = Callable[[Function], None]

class BaseMemory(BaseModel, IMemory[Role, AbstractMessage]):
    """
    The Memory class provides storage and handling of messages for a language model session.
    It supports adding, removing, and handling different types of messages and
    function calls categorized by roles: assistant, tool, user, and system.
    """

    messages: List[AbstractMessage] = Field(default_factory=list, alias='messages')

    # Private dict of message handlers for each role
    _on_message: Dict[Role, List[OnMessageCallable]] = PrivateAttr(default_factory=lambda: {
        Role.assistant: [],
        Role.tool: [],
        Role.user: [],
        Role.system: [],
    })

    def handle_message(self, message: AbstractMessage) -> None:
        """
        Implements the abstract method to handle messages based on their roles.
        """
        role: Optional[Role] = message.get("role")
        if role is None:
            raise ValueError("Message does not have a role")
        for handler in self._on_message.get(role, []):
            handler(message)


    # Overriding add_message with specific implementations
    @singledispatchmethod
    def add_message(self, message: SupportedMessages) -> None:
        """
        Overrides the abstract method and provides dispatching to specific handlers.
        see @add_message.register for the specific implementations
        """
        raise TypeError(f"Unsupported message format: {type(message)}")

    @add_message.register
    def _add_abstract_message(self, message: dict) -> None:
        """
        Handles AbstractMessage instances.
        """
        self.messages.append(message)
        self.handle_message(message)

    @add_message.register
    def _add_message_str(self, message: str) -> None:
        """
        Handles string messages.
        """
        self.add_message({"role": Role.user, "content": message})

    @add_message.register
    def _add_message_list(self, messages: list) -> None:
        """
        Handles lists of messages.
        """
        self.add_messages(messages)

    # Methods to add messages of specific roles
    def add_system_message(self, prompt: str) -> None:
        self.add_message({"role": Role.system, "content": prompt})

    def add_user_message(self, prompt: str) -> None:
        self.add_message({"role": Role.user, "content": prompt})

    def last_message_str(self) -> Optional[str]:
        message_str = None
        last_message = self.last_message
        result = last_message["content"] if "content" in last_message else last_message
        return str(result)

    def add_on_tool_call(self, fun: OnFunctionCallable) -> None:
        """
        Adds a handler to track function calls.
        """

        def tool_handler(message: AbstractMessage) -> None:
            tool_calls = message.get('tool_calls', [])
            for call in tool_calls:
                function_name = call.get('function')
                if function_name:
                    fun(function_name)
                else:
                    raise ValueError("Function name is None")

        self.add_on_message_handler(Role.assistant, tool_handler)


    def add_on_tool_message(self, handler: OnMessageCallable) -> None:
        """
        Adds a handler to be called for tool messages.

        :param handler: The callable to be executed when a tool message is added.
        """
        self.add_on_message_handler(Role.tool, handler)

    def add_on_user_message(self, handler: OnMessageCallable) -> None:
        """
        Adds a handler to be called for user messages.
        
        :param handler: The callable to be executed when a user message is added.
        """
        self.add_on_message_handler(Role.user, handler)

    def add_on_assistant_message(self, handler: OnMessageCallable) -> None:
        """
        Adds a handler to be called for assistant messages.
        
        :param handler: The callable to be executed when an assistant message is added.
        """
        self.add_on_message_handler(Role.assistant, handler)

    def add_on_system_message(self, handler: OnMessageCallable) -> None:
        """
        Adds a handler to be called for system messages.
        
        :param handler: The callable to be executed when a system message is added.
        """
        self.add_on_message_handler(Role.system, handler)


    def remove_on_tool_message(self, handler: OnMessageCallable) -> None:
        """
        Removes a specific handler for tool messages.
        
        :param handler: The handler to be removed.
        """
        self.remove_on_message_handler(Role.tool, handler)

    def remove_on_user_message(self, handler: OnMessageCallable) -> None:
        """
        Removes a specific handler for user messages.
        
        :param handler: The handler to be removed.
        """
        self.remove_on_message_handler(Role.user, handler)

    def remove_on_assistant_message(self, handler: OnMessageCallable) -> None:
        """
        Removes a specific handler for assistant messages.
        
        :param handler: The handler to be removed.
        """
        self.remove_on_message_handler(Role.assistant, handler)

    def remove_on_system_message(self, handler: OnMessageCallable) -> None:
        """
        Removes a specific handler for system messages.
        
        :param handler: The handler to be removed.
        """
        self._remove_on_message(handler, Role.system)

    def deepcopy(self) -> 'BaseMemory':
        return self.model_copy(deep=True)






import json

from litellm import ModelResponse, CustomStreamWrapper, completion, acompletion, stream_chunk_builder
from typing import Optional, Callable, Union, Coroutine, ClassVar, Type, Sequence, List, Any, AsyncGenerator

from pydantic import Field, AliasPath, PrivateAttr, BaseModel, Json, field_validator

from just_agents.core.types import AbstractMessage

from just_agents.streaming.protocols.interfaces.IFunctionCall import IFunctionCall, ToolByNameCallback
from just_agents.streaming.protocols.interfaces.IProtocolAdapter import IProtocolAdapter, ExecuteToolCallback
from just_agents.streaming.protocols.abstract_protocol import AbstractStreamingProtocol
from just_agents.streaming.protocols.openai_streaming import OpenaiStreamingProtocol


class OAIFunctionCall(BaseModel, IFunctionCall[AbstractMessage], extra="allow"):
    id: str = Field(...)
    name: str = Field(..., validation_alias=AliasPath('function', 'name'))
    arguments: Json[dict] = Field(..., validation_alias=AliasPath('function', 'arguments'))
    type: Optional[str] = Field('function')

    @classmethod
    @field_validator('arguments', mode='before')
    def stringify_arguments(cls, value):
        # Convert dict to JSON string if necessary
        if isinstance(value, dict):
            return json.dumps(value)
        return value  # Assume it's already a string

    def execute_function(self, call_by_name: ToolByNameCallback):
        try:
            function_to_call = call_by_name(self.name)
            function_args = self.arguments
            function_response = str(function_to_call(**function_args))
        except Exception as e:
            function_response = str(e)
        message = {"role": "tool", "content": function_response, "name": self.name, "tool_call_id": self.id}
        return message

    @staticmethod
    def reconstruct_tool_call_message(calls: Sequence['OAIFunctionCall']) -> dict:
        tool_calls = []
        for call_params in calls:
            tool_calls.append({"type": "function",
                               "id": call_params.id, "function": {"name": call_params.name, "arguments": str(call_params.arguments)}})
        return {"role": "assistant", "content": None, "tool_calls": tool_calls}


class OAIAdapter(BaseModel, IProtocolAdapter[ModelResponse,AbstractMessage]):
    #Class that describes function convention
    function_convention: ClassVar[Type[IFunctionCall[AbstractMessage]]] = OAIFunctionCall
    #hooks to agent class
    execute_function_hook:  ExecuteToolCallback[AbstractMessage] = Field(...)
    _output_streaming: AbstractStreamingProtocol = PrivateAttr(default_factory=OpenaiStreamingProtocol)

    def model_post_init(self, __context: Any) -> None:
        super().model_post_init(__context)

    def completion(self, *args, **kwargs) -> ModelResponse:
        return completion(*args, **kwargs)

    async def async_completion(self, *args, **kwargs) \
            -> Coroutine[Any, Any, Union[ModelResponse, CustomStreamWrapper, AsyncGenerator]]:
        return acompletion(*args, **kwargs)

    def message_from_response(self, response: ModelResponse) -> AbstractMessage:
        message = response.choices[0].message.model_dump(
            mode="json",
            exclude_none=True,
            exclude_unset=True,
            by_alias=True,
            exclude={"function_call"} if not response.choices[0].message.function_call else {}  # failsafe
        )
        if "citations" in response:
            message["citations"] = response.citations #perplexity specific field
        assert "function_call" not in message
        return message

    def message_from_delta(self, response: ModelResponse) -> AbstractMessage:
        message = response.choices[0].delta.model_dump(
            mode="json",
            exclude_none=True,
            exclude_unset=True,
            by_alias=True,
            exclude={"function_call"} if not response.choices[0].delta.function_call else {}  # failsafe
        )
        assert "function_call" not in message
        return message

    def content_from_delta(self, delta: AbstractMessage) -> str:
        return delta.get("content")

    def tool_calls_from_message(self, message: AbstractMessage) -> List[OAIFunctionCall]:
        # If there are no tool calls or tools available, exit the loop
        tool_calls = message.get("tool_calls")
        if not tool_calls:
            return []
        else:
            # Auto-convert each item in tool_calls to a FunctionCall instance with validation
            return [
                OAIFunctionCall(**tool_call)
                for tool_call in tool_calls
            ]

    def response_from_deltas(self, chunks: List[Any]) -> ModelResponse:
        return stream_chunk_builder(chunks)




from typing import Any, Dict, List, Optional
from pydantic import Field, HttpUrl

from just_agents.core.types import ModelOptions

LLMOptions = Dict[str, Any]

class LLMOptionsBase(ModelOptions, extra="allow"):
    api_key: Optional[str] = Field(None, examples=["sk-proj-...."])
    api_base : Optional[HttpUrl] = Field(default=None,
        examples=[
            "https://api.groq.com/openai/v1",
            "https://api.openai.com/v1"
        ])
    tools : Optional[List[Any]] = None
    tool_choice : Optional[str] = None


ANTHROPIC_CLAUDE_3_5_SONNET: dict[str, Any] = {
    "model": "claude-3-5-sonnet-20240620",
    "temperature": 0.0
}

LLAMA3_2_VISION: dict[str, Any] = {
    #supports both text and vision
    "model": "groq/llama-3.2-90b-vision-preview",
    "api_base": "https://api.groq.com/openai/v1",
    "temperature": 0.0,
    "tools": []
}


MISTRAL_8x22B = {
    "model": "mistral/open-mixtral-8x22b",
    "temperature": 0.0
}

OPENAI_GPT4o: dict[str, Any] = {
    "model": "gpt-4o",
    "temperature": 0.0
}

OPENAI_GPT4oMINI: dict[str, Any] = {
    "model": "gpt-4o-mini",
    "temperature": 0.0
}

OPENAI_O1_MINI: dict[str, Any] = {
    "model": "o1-mini",
    "temperature": 0.0
}

OPENAI_O1_PREVIEW: dict[str, Any] = {
    "model": "o1-preview",
    "temperature": 0.0
}

PERPLEXITY_LLAMA_3_1_SONAR_LARGE_128K_ONLINE: dict[str, Any]= {
    "model": "perplexity/llama-3.1-sonar-large-128k-online",
    "temperature": 0.0,
    "return_citations": True,
    "return_related_questions": True
}

OPEN_ROUTER_Qwen_2_72B_Instruct: dict[str, Any] = {
    "model": "openrouter/qwen/qwen-2-72b-instruct",
    "temperature": 0.0,
    "tools": []
}

OPEN_ROUTER_Qwen_2_72B_Instruct_Vision: dict[str, Any] = {
    "model": "openrouter/qwen/qwen-2-vl-72b-instruct",
    "temperature": 0.0,
    "tools": []
}

OPEN_ROUTER_LLAMA_3_8B_FREE: dict[str, Any] = {
    "model": "openrouter/meta-llama/llama-3-8b-instruct:free",
    "temperature": 0.0,
    "tools": []
}

OPEN_ROUTER_GEMINI_1_5_FLASH_EXP_FREE: dict[str, Any] = {
    "model": "openrouter/google/gemini-flash-1.5-exp",
    "temperature": 0.0,
    "tools": []
}


from pydantic import Field, PrivateAttr
from typing import Optional, List, Union, Any, Generator

from just_agents.core.interfaces.IMemory import IMemory
from just_agents.core.types import Role, AbstractMessage, SupportedMessages, SupportedMessage

from just_agents.llm_options import LLMOptions
from just_agents.streaming.protocols.interfaces.IFunctionCall import IFunctionCall
from just_agents.streaming.protocols.interfaces.IProtocolAdapter import IProtocolAdapter, BaseModelResponse
from just_agents.core.interfaces.IAgent import IAgentWithInterceptors, QueryListener, ResponseListener

from just_agents.base_memory import BaseMemory
from just_agents.just_profile import JustAgentProfile
from just_agents.core.rotate_keys import RotateKeys
from just_agents.streaming.protocol_factory import StreamingMode, ProtocolAdapterFactory
from litellm.litellm_core_utils.get_supported_openai_params import get_supported_openai_params


class BaseAgent(
    JustAgentProfile,
    IAgentWithInterceptors[
        SupportedMessages, #Input
        SupportedMessages, #Output
        SupportedMessages #StreamingOutput
    ]
):
    """
    A base agent that can query and stream LLM inputs and outputs.

    Note: it is based on pydantic and the only required field is llm_options.
    However, it is also recommended to set system_prompt.
    """

    # Core configuration for the LLM
    llm_options: LLMOptions = Field(
        ...,  # ... means this field is required
        validation_alias="options",
        description="options that will be passed to the LLM, see https://platform.openai.com/docs/api-reference/completions/create for more details")
    
    # Fallback options if primary LLM call fails
    backup_options: Optional[LLMOptions] = Field(
        default=None,
        exclude=True,
        description="options that will be used after we give up with main options, one more completion call will be done with backup options")

    # API key management settings
    completion_remove_key_on_error: bool = Field(
        default=True,
        description="In case of using list of keys removing key from the list after error call with this key")
    completion_max_tries: Optional[int]  = Field(
        2, ge=0,
        description="Maximum retry attempts before failing or falling back to backup_options")
    
    # Memory system to store conversation history
    memory: BaseMemory = Field(
        default_factory=BaseMemory,
        exclude=True,
        description="Stores conversation history and maintains context between messages")

    # Private attributes for internal state management
    _protocol: Optional[IProtocolAdapter] = PrivateAttr(None)  # Handles LLM-specific message formatting
    _partial_streaming_chunks: List[BaseModelResponse] = PrivateAttr(default_factory=list)  # Buffers streaming responses
    _key_getter: Optional[RotateKeys] = PrivateAttr(None)  # Manages API key rotation

    streaming_method: StreamingMode = Field(
        StreamingMode.openai,
        description="protocol to handle llm format for function calling")

    key_list_path: Optional[str] = Field(
        default=None,
        exclude=True,
        description="path to text file with list of api keys, one key per line")
    drop_params: bool = Field(
        default=True,
        description=" drop params from the request, useful for some models that do not support them")

    _on_query : List[QueryListener] = PrivateAttr(default_factory=list)
    _on_response : List[ResponseListener] = PrivateAttr(default_factory=list)

    memory: BaseMemory = Field(default_factory=BaseMemory,
        exclude=True,
        description="memory for the agent that stores previous messages") #is supposed to be IMemory[Role,SupportedMessage]
    
    _protocol: Optional[IProtocolAdapter] = PrivateAttr(None)
    _partial_streaming_chunks: List[BaseModelResponse] = PrivateAttr(default_factory=list)
    _key_getter: Optional[RotateKeys] = PrivateAttr(None)


    def instruct(self, prompt: str): #backward compatibility
        self.memory.add_message({"role": Role.system, "content": prompt})

    def clear_memory(self) -> None:
        self.memory.clear_messages()
        self.instruct(self.system_prompt)

    def deepcopy_memory(self) -> IMemory:
        return self.memory.deepcopy()

    def add_to_memory(self, messages: SupportedMessages) -> None:
        self.memory.add_message(messages)

    def get_last_message(self) -> SupportedMessage:  # type: ignore
        msg = self.memory.last_message
        if msg is None:
            raise ValueError("No messages in memory")
        return msg

    def model_post_init(self, __context: Any) -> None:
        # Call parent class's post_init first (from JustAgentProfile)
        super().model_post_init(__context)

        # Initialize protocol adapter if not already set
        # Protocol adapter handles formatting messages for specific LLM providers
        if not self._protocol:
            self._protocol = ProtocolAdapterFactory.get_protocol_adapter(
                self.streaming_method,  # e.g., OpenAI, Azure, etc.
                execute_functions=lambda calls: self._process_function_calls(calls),
            )

        # If tools (functions) are defined, configure LLM to use them
        if self.tools is not None:
            # Enable automatic tool selection if not explicitly set
            if not self.llm_options.get("tool_choice", None):
                self.llm_options["tool_choice"] = "auto"

        # Set up API key rotation if a key list file is provided
        if self.key_list_path is not None:
            self._key_getter = RotateKeys(self.key_list_path)
        # Warn if both direct API key and key rotation are configured
        if (self._key_getter is not None) and (self.llm_options.get("api_key", None) is not None):
            print("Warning api_key will be rewritten by key_getter. Both are present in llm_options.")

        # Initialize the agent with its system prompt
        self.instruct(self.system_prompt)

    def _prepare_options(self, options: LLMOptions):
        opt = options.copy()
        if self.tools is not None:  # populate llm_options based on available tools
            opt["tools"] = [{"type": "function",
                             "function": self.tools[tool].get_litellm_description()} for tool in self.tools]
        return opt

    def _execute_completion(
            self,
            stream: bool,
            **kwargs
    ) -> Union[AbstractMessage, BaseModelResponse]:
        
        opt = self._prepare_options(self.llm_options)
        opt.update(kwargs)
        
        max_tries = self.completion_max_tries or 1  # provide default if None
        if self._key_getter is not None:
            if max_tries < 1:
                max_tries = self._key_getter.len()
            else:
                if self.completion_remove_key_on_error:
                    max_tries = min(max_tries, self._key_getter.len())
            last_exception = None
            for _ in range(max_tries):
                opt["api_key"] = self._key_getter()
                try:
                    return self._protocol.completion(messages=self.memory.messages, stream=stream, **opt)
                except Exception as e:
                    last_exception = e
                    if self.completion_remove_key_on_error:
                        self._key_getter.remove(opt["api_key"])

            if self.backup_options:
                opt = self._prepare_options(self.backup_options)
                return self._protocol.completion(messages=self.memory.messages, stream=stream, **opt)
            if last_exception:
                raise last_exception
            else:
                raise Exception(
                    f"Run out of tries to execute completion. Check your keys! Keys {self._key_getter.len()} left.")
        else:
            return self._protocol.completion(messages=self.memory.messages, stream=stream, **opt)


    def _process_function_calls(self, function_calls: List[IFunctionCall[AbstractMessage]]) -> SupportedMessages:
        messages: SupportedMessages = []
        for call in function_calls:
            msg = call.execute_function(lambda function_name: self.tools[function_name].get_callable())
            self.handle_on_response(msg)
            self.add_to_memory(msg)
            messages.append(msg)
        return messages
    
    def query_with_currentmemory(self, **kwargs): #former proceed() aka llm_think()
        while True:
            # individual llm call, unpacking the message, processing handlers
            response = self._execute_completion(stream=False, **kwargs)
            msg: AbstractMessage = self._protocol.message_from_response(response) # type: ignore
            self.handle_on_response(msg)
            self.add_to_memory(msg)

            if not self.tools:
               break
            # If there are no tool calls or tools available, exit the loop
            tool_calls = self._protocol.tool_calls_from_message(msg)
            if not tool_calls:
                break
            # Process each tool call if they exist and re-execute query
            self._process_function_calls(tool_calls) #NOTE: no kwargs here as tool calls might need different parameters


    def streaming_query_with_currentmemory(self, reconstruct = False, **kwargs):
        try:
            self._partial_streaming_chunks.clear()
            while True: #TODO rewrite this super-ugly while-True-break stuff into proper recursion
                response = self._execute_completion(stream=True, **kwargs)
                tool_messages: list[AbstractMessage] = []
                for i, part in enumerate(response):
                    self._partial_streaming_chunks.append(part)
                    msg: AbstractMessage = self._protocol.message_from_delta(response) # type: ignore
                    delta = self._protocol.content_from_delta(msg)
                    if delta:
                        if reconstruct:
                            yield self._protocol.get_chunk(i, delta, options={'model': part["model"]})
                        else:
                            yield response
                    if self.tools:
                        tool_calls = self._protocol.tool_calls_from_message(msg)
                        if tool_calls:
                            self.add_to_memory(
                                self._protocol.function_convention.reconstruct_tool_call_message(tool_calls)
                            )
                            self._process_function_calls(tool_calls)
                            tool_messages.append(self._process_function_calls(tool_calls))
                if not tool_messages:
                    break

        finally:
            yield self._protocol.done()
            if len(self._partial_streaming_chunks) > 0:
                response = self._protocol.response_from_deltas(self._partial_streaming_chunks)
                msg: AbstractMessage = self._protocol.message_from_response(response) # type: ignore
                self.handle_on_response(msg)
                self.add_to_memory(msg)
            self._partial_streaming_chunks.clear()


    def query(self, query_input: SupportedMessages, **kwargs) -> str:
        """
        Query the agent and return the last message
        """
        self.handle_on_query(query_input)
        self.add_to_memory(query_input)
        self.query_with_currentmemory(**kwargs)
        result = self.memory.last_message_str()
        if result is None:
            raise ValueError("No response generated")
        return result
    

    def stream(self, query_input: SupportedMessages, reconstruct = False, **kwargs) -> Generator[Union[BaseModelResponse, AbstractMessage],None,None]:
        self.handle_on_query(query_input)
        self.add_to_memory(query_input)
        # Pass kwargs to streaming_query_with_currentmemory
        return self.streaming_query_with_currentmemory(reconstruct=reconstruct, **kwargs)


    @property
    def model_supported_parameters(self) -> list[str]:
        """Returns the list of parameters supported by the current model"""
        model = self.llm_options.get("model")
        if not model:
            return []
        return get_supported_openai_params(model)  # type: ignore
    
    
    @property
    def supports_response_format(self) -> bool:
        """Checks if the current model supports the response_format parameter"""
        #TODO: implement provider specific check
        return "response_format" in self.model_supported_parameters


import yaml
import importlib
from typing import Optional, Dict, Any, ClassVar, Sequence, Union, Set
from pathlib import Path
from pydantic import BaseModel, Field, field_validator, ValidationError
from collections.abc import MutableMapping, MutableSequence

class JustYaml:
    """
    A utility static class for reading and saving data to YAML files.

    Methods:
        read_yaml_data(file_path: Path, section_name: str, parent_section: str = DEFAULT_AGENT_PROFILES_SECTION) -> Dict:
            Reads data from a given section within a YAML file.

        save_to_yaml(file_path: Path, section_data: Dict, section_name: str, parent_section: str = DEFAULT_AGENT_PROFILES_SECTION) -> None:
            Updates a section within a YAML file with new data.
    """

    @staticmethod
    def read_yaml_data(
            file_path: Path,
            section_name: str,
            parent_section: str = None
    ) -> Dict:
        """
        Reads data from a given section within a YAML file.

        Args:
            file_path (Path): The path to the YAML file.
            section_name (str): The name of the section to read.
            parent_section (str): The parent section name. Defaults to 'agent_profiles'.

        Returns:
            Dict: The data from the specified section.

        Raises:
            ValueError: If the section or parent section is not found in the YAML file.
        """
        if file_path.exists():
            with file_path.open('r') as f:
                data = yaml.safe_load(f) or {}
        else:
            raise FileNotFoundError(
                f"File '{file_path}' not found."
            )
        try:
            # Retrieve the data for the specified section
            if parent_section:
                return data[parent_section][section_name]
            else:
                return data[section_name]
        except KeyError:
            raise ValueError(
                f"Section '{section_name}' under parent section '{parent_section}' not found in '{file_path}'"
            )

    @staticmethod
    def read_yaml_data_safe(
            file_path: Path,
            section_name: str,
            parent_section: str = None
    ) -> Optional[Dict[str, Any]]:
        """
        Reads data from a given section within a YAML file.

        Args:
            file_path (Path): The path to the YAML file.
            section_name (str): The name of the section to read.
            parent_section (str): The parent section name. Defaults to 'agent_profiles'.

        Returns:
            Optional[Dict]: The data from the specified section, or None if not found or error occurred while reading.
        """
        if file_path.exists():
            with file_path.open('r') as f:
                data = yaml.safe_load(f) or {}
        else:
            return None
        try:
            # Retrieve the data for the specified section
            if parent_section:
                if parent_section in data:
                    if section_name and section_name in data[parent_section]:
                        return data[parent_section][section_name]
            elif section_name and section_name in data:
                return data[section_name]
            elif not section_name and data:
                return data
            else:
                return None
        except KeyError:
            return None

    @staticmethod
    def save_to_yaml(
            file_path: Path,
            section_data: Dict,
            section_name: str,
            parent_section: str = None
    ) -> None:
        """
        Updates a section within a YAML file with new data.

        Args:
            file_path (Path): The path to the YAML file.
            section_data (Dict): The data to be saved in the specified section.
            section_name (str): The name of the section to update.
            parent_section (str): The parent section name.

        Returns:
            None
        """

        data = {}
        # Check if the YAML file exists and load existing data
        if file_path.exists():
            with file_path.open('r') as f:
                existing_data = dict(yaml.safe_load(f) or {})
                data.update(existing_data)
        else:
            file_path.parent.mkdir(parents=True, exist_ok=True)

        # Ensure the parent section exists
        if parent_section:
            if parent_section not in data:
                data[parent_section] = {}

        # Update the data with the current instance's data
        if parent_section:
            data[parent_section][section_name] = section_data
        else:
            data[section_name] = section_data

        # Write the updated data back to the YAML file
        with file_path.open('w') as f:
            yaml.safe_dump(data, f)


class JustSerializable(BaseModel, extra="allow", use_enum_values=True, validate_assignment=True, populate_by_name=True):
    """
    Pydantic2 wrapper class that implements semi-automated YAML and JSON serialization and deserialization

    Constants:
        DEFAULT_CONFIG_PATH (Path): Default path to the configuration YAML file.
        DEFAULT_PARENT_SECTION (str): Default parent section name.
        DEFAULT_SECTION_NAME (str): Default section name to use when none is provided.

    """
    DEFAULT_CONFIG_PATH : ClassVar[Path] = Path('config/just_agents.yaml')
    DEFAULT_PARENT_SECTION : ClassVar[Optional[str]] = None
    DEFAULT_SECTION_NAME : ClassVar[Optional[str]] = 'RenameMe'
    #MODULE_DIR : ClassVar[Path] = Path(os.path.abspath(os.path.dirname(__file__)))

    config_path : Optional[Path] = Field(None,exclude=True)
    config_parent_section: Optional[str] = Field(None,exclude=True)

    shortname: str = Field(
        DEFAULT_SECTION_NAME,
        description="A short name used as a section name in YAML storage. Must be alphanumeric, underscores, or dashes.",
        alias="name",
        exclude=True)
    """A short name used as a section name in YAML storage. Must be alphanumeric, underscores, or dashes."""

    class_hierarchy: str = Field(
        None,
        exclude=True,
        description="Information about object's parent classes.")
    """Information about object's parent classes."""

    class_qualname: str = Field(
        None,
        description="A marker to discern versions of this class when loading from configuration YAML files")
    """A marker to discern versions of this class when loading from configuration YAML files."""

    extras: Optional[dict] = Field(
        None,
        exclude=True)
    """Fallback container to store fields that don't fit the model."""

    def model_post_init(self, __context: Any) -> None:
        """
        Custom initializer to handle extra fields by storing them in the 'extras' attribute.
        """
        super().model_post_init(__context)
        if self.class_qualname is None:
            self.class_qualname = self.get_full_class_path()
        if self.class_hierarchy is None:
            self.class_hierarchy = self.get_class_hierarchy()
        if self.shortname == self.DEFAULT_SECTION_NAME:
            self.shortname = self.__class__.__name__
        # Collect extra fields into 'extras' attribute
        extra = getattr(self, '__pydantic_extra__', {}) or {}
        if self.extras is None:
            self.extras = extra
        else:
            self.extras.update(extra)

    @classmethod
    def get_full_class_path(cls) -> str:
        """
        Returns the full class path including the module and the qualified name of the class.
        """
        return f"{cls.__module__}.{cls.__qualname__}"

    @classmethod
    def get_class_hierarchy(cls) -> str:
        """
        Extracts the class hierarchy up to the root and represents it as a string.

        Args:
            cls: The class for which to extract the inheritance chain.

        Returns:
            str: The inheritance chain as a string, e.g., "root.BaseModel.ClassA.ClassC".
        """
        # Initialize the hierarchy list with the name of the current class
        hierarchy = []
        current_class = cls

        while current_class is not None:
            hierarchy.append(current_class.__name__)
            if current_class.__bases__:
                current_class = current_class.__bases__[0]
            else:
                current_class = None

        hierarchy.reverse()
        return '.'.join(hierarchy)

    @classmethod
    @field_validator('shortname', mode="before")
    def validate_shortname(cls, value: str) -> str:
        """
        Validates the 'shortname' field to ensure it contains only allowed characters.
        Regex pattern equivalent: '/^[a-zA-Z0-9_-]+$/'
        """
        allowed_characters = set("abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789_-")
        if not set(value).issubset(allowed_characters):
            raise ValueError("Shortname can only contain alphanumeric characters, underscores, and dashes.")
        return value

    @classmethod
    def from_json(cls, json_data: Dict[str, Any], qualname_check: bool = True) -> 'JustSerializable':
        """
        Constructor from JSON data. Populates fields that are present in the class,
        and stores any extra fields in 'extras'.

        Args:
          json_data (Dict[str, Any]): The JSON data as a dictionary.
          qualname_check (bool): If True, checks if the class_qualname matches.

        Returns:
          JustSerializable: A new instance of JustYamlSerializable.

        Raises:
          ValidationError: If class_qualname does not match when qualname_check is True.
        """

        instance = cls.model_validate(json_data)
        class_qualname = cls.get_full_class_path()
        if qualname_check and class_qualname != instance.class_qualname:
            raise ValidationError(f"Field class_qualname mismatch:'{instance.class_qualname}', self:'{class_qualname}'")
        return instance

    @classmethod
    def from_yaml(cls, section_name: str,
                  parent_section: str = None,
                  file_path: Path = None,
    ) -> 'JustSerializable':
        """
        Creates an instance from a YAML file path, section name, and parent section name.

        Args:
            section_name (str): The section name (shortname) in the YAML file.
            parent_section (str): The parent section name in the YAML file.
            file_path (Path): The path to the YAML file.

        Returns:
            JustSerializable: A new instance of JustYamlSerializable.

        Raises:
            ValueError: If the specified section is not found in the YAML file.
        """
        section_data = JustYaml.read_yaml_data(
            file_path or cls.DEFAULT_CONFIG_PATH,
            section_name,
            parent_section or cls.DEFAULT_PARENT_SECTION,
        )
        if not section_data.get("config_path"):
            section_data.update({"config_path": file_path})
        if not section_data.get("config_parent_section"):
            section_data.update({"config_parent_section": parent_section})
        section_data.update({"shortname": section_name})
        return cls.model_validate(section_data)

    @staticmethod
    def from_yaml_auto(section_name: str,
                       parent_section: Optional[str],
                       file_path: Path,
                       class_hint: Optional[str] = None,
                       ) -> Any:
        """
        Creates an instance from a YAML file.

        This function reads configuration data from a specified YAML file, section name,
        and parent section name. If the configuration data contains a `class_qualname` field,
        it dynamically imports and instantiates the corresponding class. Otherwise, returns None.

        Args:
            section_name (str): The section name in the YAML file.
            parent_section (Optional[str]): The parent section name in the YAML file.
            file_path (Path): The path to the YAML file.
            class_hint (Optional[str]): Attempt instantiation with this class_qualname if not specified in schema

        Returns:
            Any: An instance of the dynamically imported class if `class_qualname` is found in the
                 configuration data; otherwise, returns None.
        """
        config_data = JustYaml.read_yaml_data_safe(
            file_path,
            section_name,
            parent_section,
        )
        if config_data is None:
            return None
        instance = None
        if not config_data.get("config_path"):
            config_data.update({"config_path": file_path})
        if not config_data.get("config_parent_section"):
            config_data.update({"config_parent_section": parent_section})
        config_data.update({"shortname": section_name})
        class_qualname = config_data.get("class_qualname") or class_hint
        if class_qualname:
            try:
                # Splits into `module.submodule` and `ClassName` for dynamic import
                module_name, class_name = class_qualname.rsplit(".", 1)
                module = importlib.import_module(module_name)
                cls = getattr(module, class_name)
                # Dynamic instantiation of `Child class` or whatever class is specified
                instance = cls.from_json(config_data)
            except Exception as e:
                raise ValueError(f"Exception occurred: {str(e)}")
            finally:
                return instance
        else:
            return None


    def to_json(
            self,
            include_extras: bool = True,
            include: Optional[Set[str]] = None,
            exclude: Optional[Set[str]] = None,
            by_alias: bool = True,
            exclude_none: bool = True,
            serialize_as_any: bool = True,
            exclude_defaults: bool = True, 
            exclude_unset: bool = True
    ) -> Dict[str, Any]:
        """
        Serializes the instance to a JSON-compatible dictionary, flattening the 'extras' back into the main dictionary.
        See pydantic documentation of the arguments, they are directly mapped to model_dump

        Args:
            include_extras (bool): Whether to exclude Extra fields of Pydantic model.
            include (Optional[Set[str]]): Set of fields to include in the output, acts as Allowlist
            exclude (Optional[Set[str]]): Set of fields to exclude from the output, acts as Deny-list
            by_alias (bool): Whether to use the field's alias (if defined) in the output.
            exclude_none (bool): Whether to exclude fields with None values from the output.
            serialize_as_any (bool): Whether to serialize values by their types.

        Returns:
            Dict[str, Any]: A dictionary representation of the instance, including extra fields.
        """
        data = self.model_dump(
            mode='json',
            by_alias=by_alias,
            exclude_none=exclude_none,
            include=include,
            exclude=exclude,
            serialize_as_any=serialize_as_any,
            exclude_defaults=exclude_defaults, 
            exclude_unset=exclude_unset
        )
        # Flatten Extras
        if include_extras and self.extras:
            data.update(self.extras)
        # Recreate dict with only public fields
        # public_data = {k: v for k, v in data.items() if not k.startswith('_')}
        return data

    def to_json_inclusive(
            self,
            include_extras: bool = True,
            include_list: Optional[Sequence[str]] = None,
            exclude_list: Optional[Sequence[str]] = None,
            by_alias: bool = True,
            exclude_none: bool = True,
            serialize_as_any: bool = True,

    ) -> Dict[str, Any]:
        """
        Serializes the instance to a JSON-compatible dictionary, changes the behavior of to override include

        Args:
            include_extras (bool): Whether to exclude Extra fields of Pydantic model.
            include_list (Optional[Sequence[str]]): List of fields to force-include in the output in addition to default set
            exclude_list (Optional[Sequence[str]]): Set of fields to exclude from the output.
            by_alias (bool): Whether to use the field's alias (if defined) in the output.
            exclude_none (bool): Whether to exclude fields with None values from the output.
            serialize_as_any (bool): Whether to serialize values by their types.

        Returns:
            Dict[str, Any]: A dictionary representation of the instance, including extra fields.
        """

        available_fields = frozenset(self.model_fields.keys())

        if exclude_list is not None:
            exclude_set = set(exclude_list) & available_fields
        else:
            exclude_set = set()

        if include_list is not None:
            include_set = set(include_list) & available_fields - exclude_set
        else:
            include_set = set()

        data = self.to_json(
            by_alias=by_alias,
            exclude_none=exclude_none,
            exclude=exclude_set,
            include_extras=include_extras,
            serialize_as_any=serialize_as_any,
        )

        for field in include_set:
            value = getattr(self, field, None)
            if value is not None or not exclude_none:
                data[field] = value

        return data

    def save_to_yaml(
            self,
            section_name: str = None,
            parent_section: str = None,
            file_path: Path = None,
            include_extras: bool = True,
            include: Optional[Set[str]] = None,
            exclude: Optional[Set[str]] = None,
            by_alias: bool = True,
            exclude_none: bool = True,
            serialize_as_any: bool = True,
            exclude_defaults: bool = True, 
            exclude_unset: bool = True
    ):
        """
        Saves the instance's data to a YAML file under the specified parent section and section name (shortname).
        If the file exists, it appends or updates the existing data.

        Args:
            section_name (str): The  section name in the YAML file, default to shortname
            parent_section (str): The parent section name in the YAML file.
            file_path (Path): The path to the YAML file.
            include_extras (bool): Whether to exclude Extra fields of Pydantic model.
            See pydantic documentation of the following flags:
            include (Optional[Set[str]]): Set of fields to include in the output, acts as Allowlist
            exclude (Optional[Set[str]]): Set of fields to exclude from the output, acts as Deny-list
            by_alias (bool): Whether to use the field's alias (if defined) in the output.
            exclude_none (bool): Whether to exclude fields with None values from the output.
            serialize_as_any (bool): Whether to serialize values by their types.
        """

        if not file_path:
            file_path = self.config_path or self.DEFAULT_CONFIG_PATH #set configured or default
        if not parent_section:
            parent_section = self.config_parent_section #None is also valid, only set configured
        if not section_name:
            section_name = self.shortname #Set configured

        section_data = self.to_json(
            include_extras=include_extras,
            include=include,
            exclude=exclude,
            by_alias=by_alias,
            exclude_none=exclude_none,
            serialize_as_any=serialize_as_any,
            exclude_defaults=exclude_defaults, 
            exclude_unset=exclude_unset
        )
        JustYaml.save_to_yaml(file_path, section_data, section_name or self.shortname, parent_section)

    def update(self, new_data: Dict, overwrite: bool = False, strict: bool = False) -> None:
        """
        Updates the fields of the current instance with the values from another JustAgentProfile instance.

        Args:
            new_data (Dict): The instance from which to update the current instance's fields.
            overwrite (bool): If True, existing fields are overwritten with new values.
            strict (bool): If True, existing fields are overwritten with new values.
        Raises:
          ValidationError: If class_qualname does not match between objects.
        """

        if new_data is None:
            return
        else:
            new_data.pop("class_qualname", None)
            #if new_qualname and new_qualname != self.class_qualname:
            #    raise ValidationError(f"Field class_qualname mismatch:'{new_qualname}', self:'{self.class_qualname}'")
            if strict:
                self.validate_keys_match(new_data)

        for field_name, field_info in self.model_fields.items():
            self_value = getattr(self, field_name, None)
            new_value : Optional[Any] = new_data.pop( field_name, None )

            if new_value is None: #there is a value to set
                continue

            if not self_value or overwrite:
                setattr(self, field_name, new_value)  #set or overwrite
            else: # self_value is set, overwrite unset, update if extension is possible
                if isinstance(self_value, MutableMapping):
                    self_value.update(new_value)
                    setattr(self, field_name, self_value)
                elif isinstance(self_value, MutableSequence):
                    self_value.extend(new_value)
                    setattr(self, field_name, self_value)
                elif isinstance(self_value, JustSerializable): #recurse
                    self_value.update(new_value.model_dump(), overwrite)
                else:
                    pass #extension not possible

        if new_data: #extra data remaining
            if self.extras is None:
                self.extras = new_data
            else:
                self.extras.update(new_data)

    def update_from_yaml(self, overwrite: bool = False):
        profile  = self.from_yaml_auto(
            self.shortname,
            parent_section=self.config_parent_section,
            file_path=self.config_path,
        )
        # Loaded some data, parameters set in init take precedence
        if profile and isinstance(profile, self.__class__):
            self.update(
                profile.to_json(
                    by_alias=False
                ),
                overwrite=overwrite
            )

    def validate_keys_match(self, instance: Union[BaseModel, Dict, Sequence]):
        """
        Dynamic validation of effective model fields compatibility between two instances.
        """
        if isinstance(instance, BaseModel):
            model_fields = set(instance.model_fields.keys())
        elif isinstance(instance, dict):
            model_fields=set(instance.keys())
        elif isinstance(instance, Sequence):
            model_fields=set(instance)
        else:
            raise TypeError(f"Unsupported type: {str(instance)}")
        this_fields = set(self.model_fields.keys())

        if model_fields.issubset(this_fields):
            missing_in_this = model_fields - this_fields
            missing_in_instance = this_fields - model_fields
            raise ValueError(f"Mismatch between model fields! Missing in Self: {missing_in_this}. In instance: {missing_in_instance}.")

    def fields_to_populate(
            self,
            include_nonnull: bool = False,
            include_list: Optional[Set[str]] = None,
            exclude_list: Optional[Set[str]] = None,
            extra_list: Dict[str, str] = None
    ) -> Dict[str, str]:
        """
        Lists the fields of the model that are suitable for being populated by language models.

        Args:
            include_nonnull (bool): If True, include fields that are already set (non-null) in the output.
            include_list (Optional[Set[str]]): A set of field names to force include in the output, when Description is available for the field.
            exclude_list (Optional[Set[str]]): A set of field names to additionally exclude from the output, include_list takes precedence.
            extra_list (Dict[str, str]): Extra fields to populate, with field names as keys and descriptions as values.

        Returns:
            Dict[str, str]: A dictionary of fields to populate, with descriptions as values.
        """
        # Set of settable fields with descriptions
        available_fields = {
            field_name for field_name, field_info in self.model_fields.items()
            if field_info.description and not field_info.frozen
        }

        excluded_fields = {
            field_name for field_name, field_info in self.model_fields.items()
            if field_info.exclude
        }

        # Extend exclude set of field names that are already set (nonnull values) if include_nonnull is False
        if not include_nonnull:
            excluded_fields.update(set({
                field_name for field_name in available_fields
                if getattr(self, field_name, None) is not None
            }))

        # Exclude set of field names that were explicitly marked for exclusion
        if exclude_list:
            excluded_fields.update(set(exclude_list))

        # Reduce exclude set by field names that are in include list
        if include_list:
            excluded_fields.difference_update(set(include_list))

        # Calculate eligible fields: all described fields minus frozen, nonnull (if applicable), excluded, except included
        eligible_fields = (available_fields - excluded_fields)

        # Create the final field_list dictionary using the eligible fields
        fields_to_populate = {field_name: self.model_fields[field_name].description for field_name in eligible_fields}

        # Top up with extras if requested
        if extra_list:
            for field, value in list(extra_list.items()):
                if field not in self.extras:
                    fields_to_populate[field] = value

        return fields_to_populate




from pathlib import Path
from pydantic import Field, model_validator
from typing import Optional, List, ClassVar, Tuple, Sequence, Callable, Dict
from just_agents.core.just_serialization import JustSerializable
from just_agents.core.just_tool import JustTool, JustTools


class JustAgentProfile(JustSerializable):
    """
    A Pydantic model representing an agent profile
    """
    DEFAULT_GENERIC_PROMPT: ClassVar[str] = "You are a helpful AI assistant"
    DEFAULT_PARENT_SECTION: ClassVar[str] = 'agent_profiles'
    DEFAULT_CONFIG_PATH: ClassVar[Path] = Path('config/agent_profiles.yaml')
    config_parent_section: Optional[str] = Field(DEFAULT_PARENT_SECTION, exclude=True)

    system_prompt: str = Field(
        DEFAULT_GENERIC_PROMPT,
        description="System prompt of the agent")
    """System prompt of the agent."""

    
    tools: Optional[JustTools] = Field(
        None,
        description="A List[Callable] of tools s available to the agent and their descriptions")
    """A List[Callable] of tools s available to the agent and their descriptions"""

    
    @model_validator(mode='after')
    def validate_model(self) -> 'JustAgentProfile':
        """Converts callables to JustTool instances and refreshes them before validation."""
        if not self.tools:
            return self
        if isinstance(self.tools, Dict):
            return self
        elif not isinstance(self.tools, Sequence):
            raise TypeError("The 'tools' field must be a sequence of callables or JustTool instances.")
        # elif not {x for x in self.tools if not isinstance(x, JustTool)}: #no items that are not JustTools
        #     return self

        new_tools = {}
        for item in self.tools:
            if isinstance(item, JustTool):
                if item.auto_refresh:
                    item=item.refresh()
                new_tools[item.name]= item
            elif callable(item):
                new_tool = JustTool.from_callable(item)
                new_tools[new_tool.name] = new_tool
            else:
                raise TypeError("Items in 'tools' must be callables or JustTool instances.")
        setattr(self, 'tools', new_tools)
        return self

    def get_tools_callables(self) -> Optional[List[Callable]]:
        """Retrieves the list of callables from the tools."""
        if self.tools is None:
            return None
        return [tool.get_callable(refresh=tool.auto_refresh) for tool in self.tools]

    @staticmethod
    def auto_load(
                section_name: str,
                parent_section: Optional[str] = DEFAULT_PARENT_SECTION,
                file_path: Path = DEFAULT_CONFIG_PATH,
        ) -> JustSerializable:
        """
        Creates an instance from a YAML file.

        This function reads configuration data from a specified YAML file, section name,
        and parent section name. If the configuration data contains a `class_qualname` field,
        it dynamically imports and instantiates the corresponding class. Otherwise, returns None.

        Args:
            section_name (str): The section name in the YAML file.
            parent_section (Optional[str]): The parent section name in the YAML file.
            file_path (Path): The path to the YAML file.

        Returns:
            Any: An instance of the dynamically imported class if `class_qualname` is found in the
                 configuration data; otherwise, returns None.
        """
        return JustSerializable.from_yaml_auto(section_name, parent_section, file_path)

    @staticmethod
    def load_legacy_schema(
            file_path: Path = DEFAULT_CONFIG_PATH,
            class_hint: str = "just_agents.base_agent.BaseAgent",
        ) -> JustSerializable:
        """
        Creates an instance from a YAML file.

        This function reads configuration data from a specified YAML file with agent schema,
        it dynamically imports and instantiates the corresponding class. Otherwise, returns None.

        Args:
            file_path (Path): The path to the YAML file.
            class_hint (str): A substitute class to use if not specified by legacy schema

        Returns:
            Any: An instance of the dynamically imported class if `class_qualname` is found in the
                 configuration data; otherwise, returns None.
        """
        return JustSerializable.from_yaml_auto("", "", file_path, class_hint=class_hint)
    

    @staticmethod
    def convert_from_legacy(
            legacy_file_path: Path,
            output_file_path: Optional[Path] = None,
            class_hint: str = "just_agents.base_agent.BaseAgent",
            section_name: Optional[str] = None,
            parent_section: Optional[str] = DEFAULT_PARENT_SECTION,
            exclude_defaults: bool = True,
            exclude_unset: bool = True
        ) -> 'JustAgentProfile':
        """
        Converts a legacy agent schema file to the new format and saves it.

        Args:
            legacy_file_path (Path): Path to the legacy schema YAML file
            output_file_path (Path): Path where the converted file should be saved
            section_name (str): Name of the section to save the converted agent under
            parent_section (str): Parent section name to save the converted agent under
            class_hint (str): A substitute class to use if not specified by legacy schema

        Returns:
            JustAgentProfile: The converted agent profile instance
        """
        # Load the legacy schema
        agent = JustAgentProfile.load_legacy_schema(
            file_path=legacy_file_path,
            class_hint=class_hint
        )

        file_path = output_file_path or legacy_file_path
        
        # Save in new format
        agent.save_to_yaml(section_name=section_name, 
                           parent_section=parent_section, 
                           file_path=file_path, 
                           exclude_defaults=exclude_defaults, 
                           exclude_unset=exclude_unset)
        
        return agent


class JustAgentFullProfile(JustAgentProfile):
    """
    A Pydantic model representing an agent profile with extended attributes.
    """

    DEFAULT_DESCRIPTION: ClassVar[str] = "Generic all-purpose AI agent"


    description: str = Field(
        DEFAULT_DESCRIPTION,
        description="Short description of what the agent does")
    """Short description of what the agent does."""

    TO_REFRESH: ClassVar[Tuple[str, ...]] = ('shortname', 'description')
    """Fields to force-renew using LLM"""

    role: Optional[str] = Field(
        None,
        description="Role of the agent")
    """Role of the agent."""

    goal: Optional[str] = Field(
        None,
        description="Goal of the agent")
    """Goal of the agent."""

    task: Optional[str] = Field(
        None,
        description="Tasks of the agent")
    """Tasks of the agent."""

    expertise_domain: Optional[str] = Field(
        None,
        description="Agent's field of expertise")
    """Agent's field of expertise."""

    limitations: Optional[str] = Field(
        None,
        description="Agent's known limitations")
    """Agent's known limitations."""

    backstory: Optional[str] = Field(
        None,
        description="Backstory of the agent")
    """Backstory of the agent."""

    llm_model_name: Optional[str] = Field(
        None,
        description="The name of the preferred model to use for inference",
        alias="model_name" #model_name conflicts with pydantic
    )
    """The name of the preferred model to use for inference"""

    knowledge_sources: Optional[List[str]] = Field(
        None,
        description="A List[str] of of external knowledge sources the agent is capable of accessing, e.g., databases, APIs, etc.")
    """List of external knowledge sources the agent is capable of accessing, e.g., databases, APIs, etc."""

    def __str__(self) -> str:
        """
        Returns the 'description' field when the instance is converted to a string.
        """
        return self.description



from typing import Callable, Optional, List, Dict, Any, Sequence, Union, Literal

from litellm.utils import function_to_dict
from pydantic import BaseModel, Field, PrivateAttr
import importlib
import inspect

FunctionParamFields=Literal["kind","default","type_annotation"]
FunctionParams = List[Dict[str, Dict[FunctionParamFields,Optional[str]]]]

class LiteLLMDescription(BaseModel, populate_by_name=True):
    name: Optional[str] = Field(..., alias='function', description="The name of the function")
    description: Optional[str] = Field(None, description="The docstring of the function.")
    parameters: Optional[Dict[str,Any]]= Field(None, description="Parameters of the function.")

class JustTool(LiteLLMDescription):
    package: str = Field(..., description="The name of the module where the function is located.")
    auto_refresh: bool = Field(True, description="Whether to automatically refresh the tool after initialization.")
    arguments: Optional[FunctionParams] = Field(
         None, description="List of parameters with their details."
    )
    _callable: Optional[Callable] = PrivateAttr(default=None)

    def model_post_init(self, __context):
        """Called after the model is initialized. Refreshes the tools metainfo if auto_refresh is True."""
        super().model_post_init(__context)
        if self.auto_refresh:
            self.refresh()

    def get_litellm_description(self) -> Dict[str, Any]:
        dump = self.model_dump(
            mode='json',
            by_alias=False,
            exclude_none=True,
            serialize_as_any=False,
            include=set(super().model_fields)
        )
        return dump

    @classmethod
    def from_callable(cls, input_function: Callable) -> 'JustTool':
        """Create a JustTool instance from a callable."""
        package = input_function.__module__
        litellm_description = function_to_dict(input_function)
        arguments = cls._extract_parameters(input_function)
        return cls(
            **litellm_description,
            package=package,
            arguments=arguments,
            _callable=input_function,
        )

    @staticmethod
    def _extract_parameters(func: Callable) -> List[Dict[str, Any]]:
        """Extract parameters from the function's signature."""
        signature = inspect.signature(func)
        parameters = []
        for name, param in signature.parameters.items():
            param_info = {
                'kind': str(param.kind),
                'default': str(param.default) if param.default != param.empty else None,
                'type_annotation': str(param.annotation) if param.annotation != param.empty else None
            }
            parameters.append({ name: param_info})
        return parameters

    def refresh(self)->'JustTool':
        """
        Refresh the JustTool instance to reflect the current state of the actual function.
        Updates package, function name, description, parameters, and ensures the function is importable.
        Returns:
            JustTool: Returns self to allow method chaining or direct appending.
        """
        try:
            # Import the module
            package = importlib.import_module(self.package)
            # Get the function from the module
            func = getattr(package, self.name)
            # Update LiteLLM description
            litellm_description = LiteLLMDescription (**function_to_dict(func))
            # Update the description
            self.description = litellm_description.description
            # Update parameters
            self.parameters= litellm_description.parameters
            self.arguments = self._extract_parameters(func)
            # Update the cached callable
            self._callable = func

            return self  # Return self to allow chaining or direct appending
        except (ImportError, AttributeError) as e:
            raise ImportError(f"Error refreshing {self.name} from {self.package}: {e}")

    def get_callable(self, refresh: bool = False) -> Callable:
        """
        Retrieve the callable function.
        If refresh is True, the callable is refreshed before returning.
        """
        if refresh:
            self.refresh()
        if self._callable is not None:
            return self._callable
        try:
            package = importlib.import_module(self.package)
            func = getattr(package, self.name)
            self._callable = func  # Cache the callable
            return func
        except (ImportError, AttributeError) as e:
            raise ImportError(f"Error importing {self.name} from {self.package}: {e}")

    def __call__(self, *args, **kwargs):
        """Allows the JustTool instance to be called like a function."""
        func = self.get_callable()
        return func(*args, **kwargs)

JustTools = Union[
    Dict[str, JustTool],  # A dictionary where keys are strings and values are JustTool instances.
    Sequence[
        Union[JustTool, Callable]
    ]  # A sequence (like a list or tuple) containing either JustTool instances or callable objects (functions).
]
# Although an internal dictionary representation is preferable, a list representation of tools can be handled and converted to a dictionary using validation.



from abc import ABC, abstractmethod
from pydantic import BaseModel
from typing import Callable, Coroutine, Union, AsyncGenerator, List, Sequence, ClassVar, Type, TypeVar, Generic, Any
from just_agents.streaming.protocols.abstract_protocol import AbstractStreamingProtocol
from just_agents.streaming.protocols.interfaces.IFunctionCall import IFunctionCall

BaseModelResponse = TypeVar('BaseModelResponse', bound=BaseModel)
AbstractMessage = TypeVar("AbstractMessage")

ModelResponseCallback=Callable[...,BaseModelResponse]
MessageUnpackCallback=Callable[[BaseModelResponse], AbstractMessage]
ExecuteToolCallback=Callable[[Sequence[IFunctionCall]],List[AbstractMessage]]

class IProtocolAdapter(AbstractStreamingProtocol, ABC, Generic[BaseModelResponse, AbstractMessage]):
    """
    Class that is required to wrap the model protocol
    """
    function_convention: ClassVar[Type[IFunctionCall[Any]]]
    _output_streaming: AbstractStreamingProtocol
    execute_function_hook: ExecuteToolCallback[AbstractMessage]

    @abstractmethod
    def completion(self, *args, **kwargs) -> BaseModelResponse:
        raise NotImplementedError("You need to implement completion first!")

    async def async_completion(self, *args, **kwargs) \
            -> Coroutine[Any,Any,Union[BaseModelResponse, AsyncGenerator, Any]]:
        raise NotImplementedError("You need to implement async_completion first!")

    @abstractmethod
    def message_from_response(self, response: BaseModelResponse) -> AbstractMessage:
        raise NotImplementedError("You need to implement message_from_response first!")

    @abstractmethod
    def message_from_delta(self, response: BaseModelResponse) -> AbstractMessage:
        raise NotImplementedError("You need to implement message_from_delta first!")

    @abstractmethod
    def content_from_delta(self, delta: AbstractMessage) -> str:
        raise NotImplementedError("You need to implement content_from_delta first!")

    @abstractmethod
    def tool_calls_from_message(self, message: AbstractMessage) -> List[IFunctionCall[AbstractMessage]]:
        raise NotImplementedError("You need to implement tool_calls_from_response first!")

    @abstractmethod
    def response_from_deltas(self, deltas: List[BaseModelResponse]) -> BaseModelResponse:
        raise NotImplementedError("You need to implement message_from_deltas first!")

    def get_chunk(self, index:int, delta:str, options:dict) -> BaseModelResponse:
        return self._output_streaming.get_chunk(index, delta, options)

    def done(self) -> str:
        return self._output_streaming.done()


from enum import Enum


from typing import Type, TypeVar, Any, List, Union, Optional, Literal, cast, TypeAlias, Sequence, Callable, Dict

from openai.types import CompletionUsage
from openai.types.chat import ChatCompletionSystemMessageParam, ChatCompletionUserMessageParam, ChatCompletionAssistantMessageParam, ChatCompletionToolMessageParam,ChatCompletionFunctionMessageParam
from openai.types.chat.chat_completion import ChatCompletion, Choice, ChatCompletionMessage
from pydantic import BaseModel, Field, HttpUrl


######### Common ###########
AbstractMessage = Dict[str, Any]
SupportedMessage = Union[str, AbstractMessage]
SupportedMessages = Union[SupportedMessage, List[SupportedMessage]]
Output = TypeVar('Output')

class Role(str, Enum):
    system = "system"
    user = "user"
    assistant = "assistant"
    tool = "tool"
    # make it similar to Literal["system", "user", "assistant", tool] while retaining enum convenience

    def __new__(cls, value, *args, **kwargs):
        obj = str.__new__(cls, value)
        obj._value_ = value
        return obj

    def __str__(self):
        return str(self.value)
    
######### Helper ###########

class ModelOptions(BaseModel):
    model: str = Field(
        ...,
        examples=["gpt-4o-mini"],
        description="LLM model name"
    )
    temperature: Optional[float] = Field(
        0.0,
        ge=0.0,
        le=2.0,
        examples=[0.7],
        description="Sampling temperature, values from 0.0 to 2.0"
    )
    top_p: Optional[float] = Field(
        1.0,
        ge=0.0,
        le=1.0,
        examples=[0.9],
        description="Nucleus sampling probability, values from 0.0 to 1.0"
    )
    presence_penalty: Optional[float] = Field(
        0.0,
        ge=-2.0,
        le=2.0,
        examples=[0.6],
        description="Presence penalty, values from -2.0 to 2.0"
    )
    frequency_penalty: Optional[float] = Field(
        0.0,
        ge=-2.0,
        le=2.0,
        examples=[0.5],
        description="Frequency penalty, values from -2.0 to 2.0"
    )